// Implemented as a red-black tree

#include <stdio.h>
#include <stdlib.h>

#include "ost.h"

enum ost_color { RED, BLACK };

struct ost_node
   {
   enum ost_color color;
   unsigned int size;
   struct ost_node * parent;
   struct ost_node * left;
   struct ost_node * right;
   void * data;
   };

struct ost
   {
   struct ost_node nil_node;
   struct ost_node * nil;
   struct ost_node * root;
   size_t size;
   bool (*le)(void * a, void * b);
   bool (*eq)(void * a, void * b);
   };

struct ost_iterator
   {
   struct ost * ost;
   struct ost_node * curr;
   };

void * ost_data_search  (struct ost * ost, struct ost_node * node, void * data);
void * ost_node_minimum (struct ost * ost, struct ost_node * node);
void * ost_node_maximum (struct ost * ost, struct ost_node * node);

void ost_data_walk_inorder   (struct ost * ost, struct ost_node * node, void (*op)(void * data));
void ost_data_walk_preorder  (struct ost * ost, struct ost_node * node, void (*op)(void * data));
void ost_data_walk_postorder (struct ost * ost, struct ost_node * node, void (*op)(void * data));
void ost_node_walk_postorder (struct ost * ost, struct ost_node * node, void (*op)(struct ost_node * n));

void ost_node_free (struct ost_node * node);

struct ost_node * ost_node_search (struct ost * ost, struct ost_node * node, void * data);

struct ost_node * ost_transplant  (struct ost * ost, struct ost_node * a, struct ost_node * b);
void ost_rotate_left  (struct ost * ost, struct ost_node * node);
void ost_rotate_right (struct ost * ost, struct ost_node * node);
void ost_insert_fixup (struct ost * ost, struct ost_node * node);
void ost_remove_fixup (struct ost * ost, struct ost_node * node);

void ost_node_dump_preorder(struct ost * ost, struct ost_node * node, void (*print_val)(void *), int depth);

////////////////////////////////////////////////////////////////////////////////
struct ost *
ost_new(bool (*le)(void * a, void * b),
	bool (*eq)(void * a, void * b))
   {
   if ( (NULL == le) || (NULL == eq) )
      return NULL;

   struct ost * ost = malloc(sizeof(struct ost));
   if (NULL == ost)
      return NULL;

   ost->nil_node = (struct ost_node) { BLACK, 0, &ost->nil_node, &ost->nil_node, &ost->nil_node, NULL };
   ost->nil = &(ost->nil_node);
   ost->root = ost->nil;
   ost->size = 0;
   ost->le = le;
   ost->eq = eq;
   return ost;
   }


////////////////////////////////////////////////////////////////////////////////
void
ost_delete(struct ost * ost)
   {
   if (NULL == ost) return;
   if (0 < ost->size)
      {
      // Walk the tree and delete the nodes.
      // This may cause memory leaks, so hopefully the user has already
      // emptied the tree.
      ost_node_walk_postorder(ost, ost->root, ost_node_free);
      }
   free(ost);
   }

////////////////////////////////////////////////////////////////////////////////
void
ost_node_free(struct ost_node * node)
   {
   free(node);
   }

////////////////////////////////////////////////////////////////////////////////
void *
ost_insert(struct ost * ost, void * data)
   {
   if ( (NULL == ost) || (NULL == data) )
      return NULL;

   struct ost_node * new = malloc(sizeof(struct ost_node));
   if (NULL == new)
      return NULL;
   new->left = new->right = new->parent = ost->nil;
   new->color = RED;
   new->size = 1;
   new->data = data;

   if (ost->nil == ost->root)
      {
      ost->root = new;
      new->color = BLACK;
      new->parent = ost->nil;
      ost->size = 1;
      return new->data;
      }

   struct ost_node * p = ost->nil;
   struct ost_node * c = ost->root;

   while (ost->nil != c)
      {
      c->size++;
      p = c;
      if (ost->le(new->data, c->data))
	 c = c->left;
      else
	 c = c->right;
      }
   new->parent = p;
   if (ost->le(new->data, p->data))
      p->left = new;
   else
      p->right = new;

   ost_insert_fixup(ost, new);

   ost->size++;
   return new->data;
   }

////////////////////////////////////////////////////////////////////////////////
void
ost_insert_fixup(struct ost * ost, struct ost_node * node)
   {
   if ( (NULL == ost) || (NULL == node) )
      return;

   while (RED == node->parent->color)
      {
      if (node->parent == node->parent->parent->left)
	 {
	 struct ost_node * y = node->parent->parent->right;
	 if (RED == y->color)
	    {
	    node->parent->color = BLACK;
	    y->color = BLACK;
	    node->parent->parent->color = RED;
	    node = node->parent->parent;
	    }
	 else 
	    {
	    if (node == node->parent->right)
	       {
	       node = node->parent;
	       ost_rotate_left(ost, node);
	       }
	    node->parent->color = BLACK;
	    node->parent->parent->color = RED;
	    ost_rotate_right(ost, node->parent->parent);
	    }
	 }
      else
	 {
	 struct ost_node * y = node->parent->parent->left;
	 if (RED == y->color)
	    {
	    node->parent->color = BLACK;
	    y->color = BLACK;
	    node->parent->parent->color = RED;
	    node = node->parent->parent;
	    }
	 else 
	    {
	    if (node == node->parent->left)
	       {
	       node = node->parent;
	       ost_rotate_right(ost, node);
	       }
	    node->parent->color = BLACK;
	    node->parent->parent->color = RED;
	    ost_rotate_left(ost, node->parent->parent);
	    }
	 }
      }
   ost->root->color = BLACK;
   }

////////////////////////////////////////////////////////////////////////////////
void
ost_walk_inorder(struct ost * ost, void (*op)(void * data))
   {
   if ( (NULL == ost) || (ost->nil == ost->root) || (NULL == op) )
      return;

   ost_data_walk_inorder(ost, ost->root, op);
   }

////////////////////////////////////////////////////////////////////////////////
void
ost_walk_postorder(struct ost * ost, void (*op)(void * data))
   {
   if ( (NULL == ost) || (ost->nil == ost->root) || (NULL == op) )
      return;

   ost_data_walk_postorder(ost, ost->root, op);
   }

////////////////////////////////////////////////////////////////////////////////
void
ost_walk_preorder(struct ost * ost, void (*op)(void * data))
   {
   if ( (NULL == ost) || (ost->nil == ost->root) || (NULL == op) )
      return;

   ost_data_walk_preorder(ost, ost->root, op);
   }

////////////////////////////////////////////////////////////////////////////////
void
ost_data_walk_inorder(struct ost * ost, struct ost_node * node, void (*op)(void * data))
   {
   if (ost->nil != node)
      {
      ost_data_walk_inorder(ost, node->left, op);
      op(node->data);
      ost_data_walk_inorder(ost, node->right, op);
      }
   }

////////////////////////////////////////////////////////////////////////////////
void
ost_data_walk_postorder(struct ost * ost, struct ost_node * node, void (*op)(void * data))
   {
   if (ost->nil != node)
      {
      ost_data_walk_postorder(ost, node->left, op);
      ost_data_walk_postorder(ost, node->right, op);
      op(node->data);
      }
   }

////////////////////////////////////////////////////////////////////////////////
void
ost_data_walk_preorder(struct ost * ost, struct ost_node * node, void (*op)(void * data))
   {
   if (ost->nil != node)
      {
      op(node->data);
      ost_data_walk_preorder(ost, node->left, op);
      ost_data_walk_preorder(ost, node->right, op);
      }
   }

////////////////////////////////////////////////////////////////////////////////
void
ost_node_walk_postorder(struct ost * ost, struct ost_node * node, void (*op)(struct ost_node * n))
   {
   if (ost->nil != node)
      {
      ost_node_walk_postorder(ost, node->left, op);
      ost_node_walk_postorder(ost, node->right, op);
      op(node);
      }
   }

////////////////////////////////////////////////////////////////////////////////
void *
ost_search(struct ost * ost, void * data)
   {
   if ( (NULL == ost) || (ost->nil == ost->root) || (NULL == data) )
      return NULL;

   return ost_data_search(ost, ost->root, data);
   }

////////////////////////////////////////////////////////////////////////////////
// Recursive
void *
ost_data_search(struct ost * ost, struct ost_node * node, void * data)
   {
   if (ost->nil == node)
      return NULL;

   if (ost->eq(data, node->data))
      return node->data;

   if (ost->le(data, node->data))
      return ost_data_search(ost, node->left, data);
   else
      return ost_data_search(ost, node->right, data);
   }

////////////////////////////////////////////////////////////////////////////////
// Iterative
void *
ost_search_iterative(struct ost * ost, void * data)
   {
   if ( (NULL == ost) || (ost->nil == ost->root) || (NULL == data) )
      return NULL;

   struct ost_node * node = ost->root;
   while (1)
      {
      if (ost->nil == node)
	 return NULL;
      
      if (ost->eq(data, node->data))
	 return node->data;
      
      if (ost->le(data, node->data))
	 node = node->left;
      else
	 node = node->right;
      }
   }

////////////////////////////////////////////////////////////////////////////////
void *
ost_minimum(struct ost * ost)
   {
   if (NULL == ost)
      return NULL;

   return ost_node_minimum(ost, ost->root);
   }

////////////////////////////////////////////////////////////////////////////////
void *
ost_node_minimum(struct ost * ost, struct ost_node * node)
   {
   while ( (ost->nil != node) && (ost->nil != node->left) )
      node = node->left;
   if (ost->nil == node)
      return NULL;
   return node->data;
   }

////////////////////////////////////////////////////////////////////////////////
void *
ost_maximum(struct ost * ost)
   {
   if (NULL == ost)
      return NULL;

   return ost_node_maximum(ost, ost->root);
   }

////////////////////////////////////////////////////////////////////////////////
void *
ost_node_maximum(struct ost * ost, struct ost_node * node)
   {
   while ( (ost->nil != node) && (ost->nil != node->right) )
      node = node->right;
   if (ost->nil == node)
      return NULL;
   return node->data;
   }

////////////////////////////////////////////////////////////////////////////////
struct ost_iterator *
ost_begin(struct ost * ost)
   {
   if ( (NULL == ost) || (ost->nil == ost->root) )
      return NULL;

   struct ost_node * min = ost->root;
   while (ost->nil != min->left)
      min = min->left;

   struct ost_iterator * it = malloc(sizeof(struct ost_iterator));
   if (NULL == it)
      return NULL;

   it->ost = ost;
   it->curr = min;
   return it;
   }

////////////////////////////////////////////////////////////////////////////////
struct ost_iterator *
ost_end(struct ost * ost)
   {
   if ( (NULL == ost) || (ost->nil == ost->root) )
      return NULL;

   struct ost_node * max = ost->root;
   while (ost->nil != max->right)
      max = max->right;

   struct ost_iterator * it = malloc(sizeof(struct ost_iterator));
   if (NULL == it)
      return NULL;

   it->ost = ost;
   it->curr = max;
   return it;
   }

////////////////////////////////////////////////////////////////////////////////
struct ost_iterator *
ost_next(struct ost_iterator * it)
   {
   if ( (NULL == it) || (NULL == it->ost) || (it->ost->nil == it->curr) )
      return NULL;

   if (it->ost->nil == it->curr->right)
      {
      struct ost_node * p = it->curr;
      it->curr = it->curr->parent;
      while ( (it->ost->nil != it->curr) && (p == it->curr->right) )
	 {
	 p = it->curr;
	 it->curr = it->curr->parent;
	 }

      if (it->curr == it->ost->nil)
	 return NULL;
      }
   else
      {
      struct ost_node * min = it->curr->right;
      while (it->ost->nil != min->left)
	 min = min->left;
      it->curr = min;
      }
   return it;
   }

////////////////////////////////////////////////////////////////////////////////
struct ost_iterator *
ost_prev(struct ost_iterator * it)
   {
   if ( (NULL == it) || (NULL == it->ost) || (it->ost->nil == it->curr) )
      return NULL;

   if (it->ost->nil == it->curr->left)
      {
      struct ost_node * p = it->curr;
      it->curr = it->curr->parent;
      while ( (it->ost->nil != it->curr) && (p == it->curr->left) )
	 {
	 p = it->curr;
	 it->curr = it->curr->parent;
	 }

      if (it->curr == it->ost->nil)
	 return NULL;
      }
   else
      {
      struct ost_node * max = it->curr->left;
      while (it->ost->nil != max->right)
	 max = max->right;
      it->curr = max;
      }
   return it;
   }

////////////////////////////////////////////////////////////////////////////////
void *
ost_value(struct ost_iterator * it)
   {
   if ( (NULL == it) || (NULL == it->ost) || (it->ost->nil == it->curr) )
      return NULL;

   return it->curr->data;
   }

////////////////////////////////////////////////////////////////////////////////
void *
ost_find(struct ost_iterator * it, void * data)
   {
   if ( (NULL == it) || (NULL == it->ost) || (it->ost->nil == it->ost->root) || (NULL == data) )
      return NULL;

   return ost_node_search(it->ost, it->ost->root, data);
   }

////////////////////////////////////////////////////////////////////////////////
// Recursive
struct ost_node *
ost_node_search(struct ost * ost, struct ost_node * node, void * data)
   {
   if (ost->nil == node)
      return NULL;

   if (ost->eq(data, node->data))
      return node;

   if (ost->le(data, node->data))
      return ost_node_search(ost, node->left, data);
   else
      return ost_node_search(ost, node->right, data);
   }

////////////////////////////////////////////////////////////////////////////////
struct ost_node *
ost_transplant(struct ost * ost, struct ost_node * a, struct ost_node * b)
   {
   if ( (NULL == a) || (NULL == b) )
      return NULL;

   if (ost->nil == a->parent)
      ost->root = b;
   else if (a == a->parent->left)
      a->parent->left = b;
   else
      a->parent->right = b;

   b->parent = a->parent;

   return a;
   }

////////////////////////////////////////////////////////////////////////////////
void *
ost_remove(struct ost_iterator * it)
   {
   if ( (NULL == it) || (NULL == it->ost) || (it->ost->nil == it->curr) )
      return NULL;

   // Get successor and save it
   struct ost_iterator succ;
   succ.curr = it->curr;
   succ.ost = it->ost;
   ost_next(&succ);

   struct ost_node * x;
   struct ost_node * y = it->curr;
   enum ost_color original_color = y->color;

   if (it->ost->nil == it->curr->left)
      {
      x = it->curr->right;
      ost_transplant(it->ost, it->curr, it->curr->right);
      }
   else if (it->ost->nil == it->curr->right)
      {
      x = it->curr->left;
      ost_transplant(it->ost, it->curr, it->curr->left);
      }
   else
      {
      y = ost_node_minimum(it->ost, it->curr->right);
      original_color = y->color;
      x = y->right;
      if (it->curr == y->parent)
	 x->parent = y;
      else
	 {
	 ost_transplant(it->ost, y, y->right);
	 y->right = it->curr->right;
	 y->right->parent = y;
	 }
      ost_transplant(it->ost, it->curr, y);
      y->left = it->curr->left;
      y->left->parent = y;
      y->color = it->curr->color;
      }

   if (BLACK == original_color)
      ost_remove_fixup(it->ost, x);

   void * data = it->curr->data;
   free(it->curr);

   // Update it to point to saved successor
   it->curr = succ.curr;
   it->ost->size--;

   return data;
   }

////////////////////////////////////////////////////////////////////////////////
void
ost_remove_fixup(struct ost * ost, struct ost_node * node)
   {
   while ( (ost->root != node) && (BLACK == node->color) )
      {
      if (node == node->parent->left)
	 {
	 struct ost_node * w = node->parent->right;
	 if (RED == w->color)
	    {
	    w->color = BLACK;
	    node->parent->color = RED;
	    ost_rotate_left(ost, node->parent);
	    w = node->parent->right;
	    }
	 if ( (BLACK == w->left->color) && (BLACK == w->right->color) )
	    {
	    w->left->color = RED;
	    node = node->parent;
	    }
	 else
	    {
	    if (BLACK == w->right->color)
	       {
	       w->left->color = BLACK;
	       w->color = RED;
	       ost_rotate_right(ost, w);
	       w = node->parent->right;
	       }
	    w->color = node->parent->color;
	    node->parent->color = BLACK;
	    w->right->color = BLACK;
	    ost_rotate_left(ost, node->parent);
	    node = ost->root;
	    }
	 }
      else
	 {
	 struct ost_node * w = node->parent->left;
	 if (RED == w->color)
	    {
	    w->color = BLACK;
	    node->parent->color = RED;
	    ost_rotate_right(ost, node->parent);
	    w = node->parent->left;
	    }
	 if ( (BLACK == w->right->color) && (BLACK == w->left->color) )
	    {
	    w->right->color = RED;
	    node = node->parent;
	    }
	 else
	    {
	    if (BLACK == w->left->color)
	       {
	       w->right->color = BLACK;
	       w->color = RED;
	       ost_rotate_left(ost, w);
	       w = node->parent->left;
	       }
	    w->color = node->parent->color;
	    node->parent->color = BLACK;
	    w->left->color = BLACK;
	    ost_rotate_right(ost, node->parent);
	    node = ost->root;
	    }
	 }
      }
   node->color = BLACK;
   }

////////////////////////////////////////////////////////////////////////////////
void
ost_rotate_left(struct ost * ost, struct ost_node * node)
   {
   if ( (NULL == ost) || (ost->nil == node) || (ost->nil == node->right) )
      return;

   struct ost_node * y = node->right;

   node->right = y->left;
   if (ost->nil != y->left)
      y->left->parent = node;

   y->parent = node->parent;
   if (ost->nil == node->parent)
      ost->root = y;
   else if (node == node->parent->left)
      node->parent->left = y;
   else
      node->parent->right = y;

   y->left = node;
   node->parent = y;

   y->size = node->size;
   node->size = node->left->size + node->right->size + 1;
   }

////////////////////////////////////////////////////////////////////////////////
void
ost_rotate_right(struct ost * ost, struct ost_node * node)
   {
   if ( (NULL == ost) || (ost->nil == node) || (ost->nil == node->left) )
      return;

   struct ost_node * y = node->left;

   node->left = y->right;
   if (ost->nil != y->right)
      y->right->parent = node;

   y->parent = node->parent;
   if (ost->nil == node->parent)
      ost->root = y;
   else if (node == node->parent->right)
      node->parent->right = y;
   else
      node->parent->left = y;

   y->right = node;
   node->parent = y;

   y->size = node->size;
   node->size = node->left->size + node->right->size + 1;
   }

////////////////////////////////////////////////////////////////////////////////
unsigned int ost_black_depth(struct ost * ost)
   {
   if ( (NULL == ost) || (NULL == ost->root) )
      return 0;

   struct ost_node * node = ost->root;
   unsigned int black_depth = 0;

   while (ost->nil != node)
      {
      if (BLACK == node->color)
	 black_depth++;
      node = node->left;
      }

   return black_depth;
   }

////////////////////////////////////////////////////////////////////////////////
void
ost_dump(struct ost * ost, void (*print_val)(void *) )
   {
   if ( (NULL == ost) || (ost->nil == ost->root) )
      return;
   ost_node_dump_preorder(ost, ost->root, print_val, 0);
   }

////////////////////////////////////////////////////////////////////////////////
void
ost_node_dump_preorder(struct ost * ost, struct ost_node * node, void (*print_val)(void *), int depth)
   {
   for (int i=0; i < depth; i++)
      printf("\t");
   printf("%-8s", (BLACK == node->color ? "black" : "red"));
   print_val(node->data);
   printf("\n");
   if (ost->nil != node->left)
      ost_node_dump_preorder(ost, node->left, print_val, depth+1);
   if (ost->nil != node->right)
      ost_node_dump_preorder(ost, node->right, print_val, depth+1);
   }

////////////////////////////////////////////////////////////////////////////////
struct ost_iterator *
ost_select(struct ost * ost, unsigned int i)
   {
   if ( (NULL == ost) || (ost->nil == ost->root) )
      return NULL;

   struct ost_iterator * it = malloc(sizeof(struct ost_iterator));
   if (NULL == it)
      return NULL;

   it->ost = ost;
   it->curr = ost->root;
   return ost_select_subtree(it, i);
   }

////////////////////////////////////////////////////////////////////////////////
struct ost_iterator *
ost_select_subtree(struct ost_iterator * it, unsigned int i)
   {
   if ( (NULL == it) || (NULL == it->ost) )
      return NULL;

   unsigned int r = it->curr->left->size + 1;
   if (i == r)
      return it;
   else
      {
      if (i < r)
	 {
	 it->curr = it->curr->left;
	 return ost_select_subtree(it, i);
	 }
      else
	 {
	 it->curr = it->curr->right;
	 return ost_select_subtree(it, i - r);
	 }
      }
   return it;
   }

////////////////////////////////////////////////////////////////////////////////
unsigned int
ost_rank(struct ost_iterator * it)
   {
   if ( (NULL == it) || (NULL == it->ost) )
      return 0;

   unsigned int r = it->curr->left->size + 1;
   struct ost_node * y = it->curr;
   while (y != it->ost->root)
      {
      if (y == y->parent->right)
	 r += y->parent->left->size + 1;
      y = y->parent;
      }
   return r;
   }

////////////////////////////////////////////////////////////////////////////////
unsigned int
ost_size (struct ost * ost)
   {
   if (NULL == ost)
      return 0;
   return ost->size;
   }

